function [out] = Solver_GMRES(in)
    % documentation of outputs:
    %   out.x                       :   numerical solution
    %   out.measure.matvec(i)    	:   number of matrix-vector products at i-th snap shot
    %   out.measure.iterres(i)   	:   norm of iteratively updated residual
    %   out.measure.trueres(i)     	:   norm of directly computed residual
    %   out.measure.runtime(i)     	:   time in seconds from start of method (excluded the time for computing the true residual)
    %   out.measure.runtime_mv(i)   :   time in seconds from start of method that is only required for matrix-vector products
    %
    % inputs:
    %   in.A_fun                    :   linear operator from C^N --> C^N
    %   in.b                        :   right-hand side, s.t. A_fun(x) = b is to be solved for x
    %   in.x0                       :   initial guess for x
    %   in.terminate.tolabs         :   terminate when |b-A_fun(x)|<=tolabs
    %   in.terminate.maxmatvec      :   terminate when out.matvec(i)>=maxmatvec
    %   in.terminate.maxruntime     :   terminate when out.runtime(i)>=maxruntime
    %   in.param.s                  :   s<<N integer cut space dimension
    %   in.param.L                  :   L<<N integer polynomial dimension
    
    fprintf('\n ---------- Using Solver GMRES ---------- \n');
    
    % init
    % % init performance measure
    myPerfMeasure    	= PerfMeasure(in,'GMRES');
    % % init system
    b                   = in.b;
    N                   = size(b,1);
    x0                 	= in.x0;
    A_fun               = @(v) myPerfMeasure.matvec(v);
    % % init storage
    m                   = in.terminate.maxmatvec;
    W                   = zeros(N,m+1);
    H                   = zeros(m+1,m);
    G                   = zeros(2,2,m);
    
    % start method
    % % start performance monitoring
    myPerfMeasure.resume();
    % % init method
    tolabs              = in.terminate.tolabs;
    r0                  = b - A_fun( x0 );
    beta                = norm( r0 );
    gamma               = beta * [1; zeros(m,1)];
    W(:,1)              = 1/beta * r0;
    opts_R              = [];
    opts_R.UT           = true;
    myPerfMeasure.read(x0,beta);
    % run loop
    for i=1:m
        % % Arnoldi step
        W(:,i+1)            = A_fun( W(:,i) );
        for j=1:i
            H(j,i)              = W(:,j)' * W(:,i+1);
            W(:,i+1)            = W(:,i+1) - H(j,i) * W(:,j);
        end
        H(i+1,i)            = norm( W(:,i+1) );
        W(:,i+1)            = 1/H(i+1,i) * W(:,i+1);
        % % apply rotations for columnwise QR of H
        % % % rotate the newly compute column of H
        for j=1:(i-1)
            H([j,j+1],i)        = G(:,:,j) * H([j,j+1],i);
        end
        % % % initialise subsequent rotation
        [Gl,Rl]             = qr( H([i,i+1],i) );
        % % % apply subsequent rotation
        % % % % on i-th column of H
        G(:,:,i)            = Gl;
        H([i,i+1],i)        = Rl;
        % % % % on gamma
        gamma([i,i+1],1)    = G(:,:,i) * gamma([i,i+1],1);
        % % check for termination
        beta                = abs(gamma(i+1,1));
        if( beta<=tolabs )
            zeta                = linsolve(H(1:i,1:i),gamma(1:i,1),opts_R);
            x                   = x0 + W(:,1:i) * zeta;
            myPerfMeasure.read(x,beta);
            break;
        end
        % << Performance measure
        myPerfMeasure.stop();
        % % compute true residual
        % % % compute x
        zeta                = linsolve(H(1:i,1:i),gamma(1:i,1),opts_R);
        x                   = x0 + W(:,1:i) * zeta;
        myPerfMeasure.read(x,beta);
        if( myPerfMeasure.isTerminate() )
            break;
        end
        myPerfMeasure.resume();
    end
    myPerfMeasure.stop();
    % write output
    out.x               = x;
    out.measure         = myPerfMeasure.getMeasure();
end