function [out] = Solver_IDRs_ref(in)
    % documentation of outputs:
    %   out.x                       :   numerical solution
    %   out.measure.matvec(i)    	:   number of matrix-vector products at i-th snap shot
    %   out.measure.iterres(i)   	:   norm of iteratively updated residual
    %   out.measure.trueres(i)     	:   norm of directly computed residual
    %   out.measure.runtime(i)     	:   time in seconds from start of method (excluded the time for computing the true residual)
    %   out.measure.runtime_mv(i)   :   time in seconds from start of method that is only required for matrix-vector products
    %
    % inputs:
    %   in.A_fun                    :   linear operator from C^N --> C^N
    %   in.b                        :   right-hand side, s.t. A_fun(x) = b is to be solved for x
    %   in.x0                       :   initial guess for x
    %   in.terminate.tolabs         :   terminate when |b-A_fun(x)|<=tolabs
    %   in.terminate.maxmatvec      :   terminate when out.matvec(i)>=maxmatvec
    %   in.terminate.maxruntime     :   terminate when out.runtime(i)>=maxruntime
    %   in.param.s                  :   s<<N integer cut space dimension
    %   in.param.L                  :   L<<N integer polynomial dimension
    
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'hU') )
        fprintf('\n ---------- Using Solver IDR(%i)ref with    recycling ---------- \n',in.param.s);
    else
        fprintf('\n ---------- Using Solver IDR(%i)ref without recycling ---------- \n',in.param.s);
    end
    
    % init
    % % init performance measure
    myPerfMeasure    	= PerfMeasure(in,['IDR(',num2str(in.param.s),')ref']);
    % % init system
    b                   = in.b;
    N                   = size(b,1);
    x                   = in.x0;
    A_fun               = @(v) myPerfMeasure.matvec(v);
    tolabs              = in.terminate.tolabs;
    % % init storage
    s                   = in.param.s;
    % % init recycling
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'tolabs2') )
        tolabs2             = in.recyclingparam.tolabs2;
        hU                  = zeros(N,s,1);
    else
        tolabs2             = inf;
        hU                  = [];
    end
    % start method
    % % start performance monitoring
    myPerfMeasure.resume();
    % % init method
    r                   = b - A_fun( x );
    beta                = norm( r );
    % % % init projectors
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'hU') )
        rng('default');
        P                	= orth( randn(N,s) );
        V(:,:,1)        	= in.recyclingparam.hU;
        for q=1:s
            V(:,q,2)         	= A_fun( V(:,q,1) );
        end
        Z               	= tril( P' * V(:,:,2) );
        myPerfMeasure.stop();
        myPerfMeasure.read(x,beta);
        myPerfMeasure.resume();
        eta                 = P' * r(:,1);
       	[x,r]               = dirRbio(V,0,Z,x,r,eta);
        beta                = norm( r );
    else
        [P,V,Z,x,r,beta,myPerfMeasure] ...
                            = Initialisation(A_fun,x,r,beta,s,tolabs,myPerfMeasure);
    end
    myPerfMeasure.read(x,beta);
    % run loop
    while(true)
        r1                  = A_fun( r );
        [tau,beta]          = StabCoeffs([r,r1],beta);
        x                   = x + tau * r;
        r                   = r - tau * r1;
        myPerfMeasure.read(x,beta);
        if( beta<=tolabs )
            break;
        end
        eta                 = P' * r;
        for q=1:s
            xi                  = Z \ eta;
            V(:,q,1)           	= V(:,:,1) * xi + tau * ( r - V(:,:,2) * xi );
            V(:,q,2)          	= A_fun( V(:,q,1) );
            Z(:,q)              = P' * V(:,q,2);
            myPerfMeasure.read(x,beta);
        end
        [x,r]               = dirRbio(V,0,Z,x,r,eta);
        beta                = norm( r );
        myPerfMeasure.read(x,beta);
        if(beta >= tolabs2)
            hU              = V(:,:,1);
            myPerfMeasure.setRecycling();
        end
        if( beta<=tolabs )
            break;
        end
        if( myPerfMeasure.isTerminate() )
            break;
        end
    end
    myPerfMeasure.stop();
    % write output
    out.x               = x;
    out.measure         = myPerfMeasure.getMeasure();
    if( isfield(in,'recyclingparam') && isfield(in.recyclingparam,'tolabs2') )
        out.recyclingparam.hU    = hU;
    end
end